
/**************************************************************************
 *                                                                        *
 * NAME:                               DATE:                              *
 *  gifread.c                           01.11.1998                        *
 *                                                                        *
 *                                                                        *
 *                                                                        *
 * DESCRIPTION:                                                           *
 *  Unit containig ripped code for reading and writing a gif file         *
 *  Gif writing supports files following the GIF87a standard. Read more   *
 *  from GIFSAVE.TXT and below.                                           *
 *                                                                        *
 *                                                                        *
 *                                                                        *
 * WRITTEN BY:                                                            *
 *  gifsave routines by Sverre H. Huseby <sverrehu@ifi.uio.no>            *
 *  gifread routines by Mark Morley <morley@camosun.bc.ca>                *
 *       Taken from his VGL graphics library. Read VGL.TXT                *
 *                                                                        *
 *  some minor changes by Yrj Kari-Koskinen <ykk@wwnet.fi>               *
 *                                                                        *
 *                                                                        *
 *                                                                        *
 * USAGE:                                                                 *
 *  To write to a GIF file include the file gifsave.c and use the         *
 *  fuction GIF_write():                                                  *
 *       void GIF_write(char *filename, int width, int height,            *
 *                      int numcolors, int bits_per_color, char *pal);    *
 *                                                                        *
 *  There should be a callback function for GIF_write named               *
 *  get_gif_pixel. Here's an example:                                     *
 *       static int get_gif_pixel(int x, int y) {                         *
 *         return vir_screen_buffer[y * 320 + x];                         *
 *       }                                                                *
 *                                                                        *
 *  To read from a GIF file include this file (gifread.c) and use the     *
 *  function GIF_read():                                                  *
 *       int GIF_read(char *file, char *buffer, char *pal,                *
 *                    int *width, int *height, char *backgr);             *
 *  If variables pal, width, height and/or are supplied, they are         *
 *  filled with wanted data. If buffer is not supplied, you still get     *
 *  the width, height, pal and/or backrg data.                            *
 *                                                                        *
 *                                                                        *
 * MORE INFORMATION:                                                      *
 *  One can reach for additional information about gif structure at:      *
 *    ftp.ncsa.uiuc.edu/misc/file.formats/graphics.formats/gif89a.doc     *
 *                                                                        *
 *                                                                        *
 *                                                                        *
 * LEGAL STUFF:                                                           *
 *  The Graphics Interchange Format(c) is the Copyright property of       *
 *  CompuServe Incorporated. GIF(tm) is a Service Mark property of        *
 *  CompuServe Incorporated.                                              *
 *                                                                        *
 *                                                                        *
 *                                                                        *
 **************************************************************************/


/*****************************************************************************
 VGLGIF.C

 GIF_read(char *file, char *buffer, char *palette,
          int *width, int *height, char *backgr);

 Routine to load a 256 color .GIF file into a memory buffer.
 If you supply a pointer to a palette, it will be filled in. If you supply
 a pointer to a width, height and/or background color variable, it will be
 filled in as well.

 Mark Morley
 morley@camosun.bc.ca

 Minor changes by Yrj Kari-Koskinen
 ykk@wwnet.fi
*****************************************************************************/

#define MAX_CODES     4096

static FILE*          fp;
static int            curr_size;
static int            clr;
static int            ending;
static int            newcodes;
static int            top_slot;
static int            slot;
static int            navail_bytes = 0;
static int            nbits_left = 0;
static unsigned char  b1;
static unsigned char  byte_buff[257];
static unsigned char* pbytes;
static unsigned char* stack;
static unsigned char* suffix;
static unsigned int*  prefix;

static unsigned long code_mask[13] = {
   0L,
   0x0001L, 0x0003L,
   0x0007L, 0x000FL,
   0x001FL, 0x003FL,
   0x007FL, 0x00FFL,
   0x01FFL, 0x03FFL,
   0x07FFL, 0x0FFFL
};

static int get_next_code(void) {
   register int  i;
   static unsigned long ret;

   if( ! nbits_left )
   {
      if( navail_bytes <= 0 )
      {
	 pbytes = byte_buff;
         navail_bytes = getc( fp );
	 if( navail_bytes )
	    for( i = 0; i < navail_bytes; ++i )
               *(byte_buff + i) = getc( fp );
      }
      b1 = *pbytes++;
      nbits_left = 8;
      --navail_bytes;
   }
   ret = b1 >> (8 - nbits_left);
   while( curr_size > nbits_left )
   {
      if( navail_bytes <= 0 )
      {
	 pbytes = byte_buff;
         navail_bytes = getc( fp );
	 if( navail_bytes )
	    for( i = 0; i < navail_bytes; ++i )
               *(byte_buff + i) = getc( fp );
      }
      b1 = *pbytes++;
      ret |= b1 << nbits_left;
      nbits_left += 8;
      --navail_bytes;
   }
   nbits_left -= curr_size;

   return( (int) (ret & *(code_mask + curr_size)) );
}

int GIF_read(char *file, char *buffer, RGB *pal,
             int *width, int *height) {
   unsigned char* sp;
   int            code, fc, oc;
   int            i;
   unsigned char  size;
   int            c;
   unsigned char  buf[1028];
   unsigned char  red;
   unsigned char  grn;
   unsigned char  blu;
   char           temp, interlaced;
   int            w, h, r, r2;

   fp = fopen( file, "rb" );
   if( !fp )
      return 10;

   // 13 tavua ji uupumaan headerista... kikkailua?

   for(i = 0; i < 256; i++) {
      red = getc(fp);
      grn = getc(fp);
      blu = getc(fp);

      if(pal) {
         pal[i].r = red >> 2;             // change from 8-bit
         pal[i].g = grn >> 2;             // to [0..63] format
         pal[i].b = blu >> 2;
      }
   }

   fread(buf, 1, 5, fp);     // image separator character = ","
                             // and image start (0,0) = 4 bytes
   fread(buf, 1, 2, fp);
   w = buf[0] + (buf[1] << 8);
   if( width )
      *width = w;

   fread(buf, 1, 2, fp);
   h = buf[0] + (buf[1] << 8);
   if( height )
      *height = h;

   if(!buffer) {
      fclose( fp );
      return 1;
   }

   temp = getc(fp);
   interlaced = (temp & 64) >> 6;   // is interlaced on?

   size = getc( fp );
   if(size < 2 || size > 9) {
      fclose( fp );
      return 12;
   }

      // let's allocate the mem'ry we need...
   stack = (unsigned char*) malloc(MAX_CODES + 1);
   if(stack == NULL)
     return 20;
   suffix = (unsigned char*) malloc(MAX_CODES + 1);
   if(suffix == NULL)
     return 21;
   prefix = (unsigned int*) malloc(sizeof(int) * (MAX_CODES + 1));
   if(prefix == NULL)
     return 22;      // not enuff' memory

   curr_size = size + 1;
   top_slot = 1 << curr_size;
   clr = 1 << size;
   ending = clr + 1;
   slot = newcodes = ending + 1;
   navail_bytes = nbits_left = 0;
   oc = fc = 0;
   sp = stack;
   while((c = get_next_code()) != ending) {

     if(c == clr) {
       curr_size = size + 1;
       slot = newcodes;
       top_slot = 1 << curr_size;

       while( (c = get_next_code()) == clr );

       if( c == ending )
         break;
       if( c >= slot )
	      c = 0;

	    oc = fc = c;
       *buffer++ = c;
       }
     else
       {
	    code = c;
	    if( code >= slot ) {
	      code = oc;
	      *sp++ = fc;
	      }
	    while( code >= newcodes ) {
	      *sp++ = *(suffix + code);
	      code = *(prefix + code);
	      }
	    *sp++ = code;
	    if( slot < top_slot ) {
	      *(suffix + slot) = fc = code;
	      *(prefix + slot++) = oc;
	      oc = c;
         }
       if( slot >= top_slot && curr_size < 12 ) {
         top_slot <<= 1;
	      ++curr_size;
	      }
       while( sp > stack ) {
  	      --sp;
         *buffer++ = *sp;
         }
       }
     }

   // and we'll remember to free the memory also.
   free(stack);
   free(suffix);
   free(prefix);
   fclose( fp );

   // if it's interlaced, we must do a trick or two

   if(interlaced == 1) {      // r = row_in_real_life, r2 = row_in_buf
     buffer -= w*h;
     char tempbuf[w*h];
     memcpy(tempbuf, buffer, w*h);

     for(r = 0, r2 = 0; r < h; r += 8, r2++) // c = column
       for(c = 0; c < w; c++)
         buffer[r * w + c] = tempbuf[r2 * w + c];
     for(r = 4; r < h; r += 8, r2++)
       for(c = 0; c < w; c++)
         buffer[r * w + c] = tempbuf[r2 * w + c];
     for(r = 2; r < h; r += 4, r2++)
       for(c = 0; c < w; c++)
         buffer[r * w + c] = tempbuf[r2 * w + c];
     for(r = 1; r < h; r += 2, r2++)
       for(c = 0; c < w; c++)
         buffer[r * w + c] = tempbuf[r2 * w + c];
     }

   return 0;
}

